/*  This file is a part of Puzzles, a MIDLET for playing puzzles.
    Copyright (C) 2004, Kostas Giannakakis

    Puzzles is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Puzzles is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Puzzles; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*
 * @(#)Board.java	1.0 04/02/18
 *
 */

package games.KnightsTour;

import games.*;
import java.util.Date;
import javax.microedition.lcdui.*;
import javax.microedition.midlet.MIDlet;
import java.io.IOException;

public class Board extends Canvas implements CommandListener
{

	MIDlet midlet;
	Display dpy;

	/** Knight's image */
	Image knightImage;

	// commands
	static final int CMD_ABOUT	= 0;
	static final int CMD_EXIT	= 1;
	static final int CMD_UNDO	= 2;
	static final int CMD_RESET	= 3;
	static final int CMD_BEST_TIME	= 4;
	static final int CMD_HELP	= 5;
	static final int CMD_ZLAST	= 6;	// must be ze last, of course
	Command cmd[];

	// state variables
	static final int START = 0;
	static final int PLAYING = 1;
	static final int WON = 2;
	static final int ENDED = 3;
	int gameState;

	// Drawing constants
	static final int TILE_SIZE = 11;
	static final int BOARD_SIZE = TILE_SIZE*8;
	static final int X_OFFSET = 4;
	static final int Y_OFFSET = 4;

	/** Position of selected square 0..55. */
	int selPos;

	/**
	* Table for storing moves. When the user makes a
	* move, they fill the next free element of movesTable with a value
	* describing the move. At the beginning the table is initialised with
	* value -1, which represents an empty element. To solve the puzzle a
	* total of movesTableSize moves must be made, filling the entire
	* table with values > -1.
	*/
	int [] movesTable;

	/** Size of movesTable.*/
	int movesTableSize = 0;

    /** Board's size.*/
    int boardSize = 8;

    /** Value of last number placed.*/
    static int currentValue;

    /** Position of last number placed.*/
    static int currentPosition;

    /**
     * Holds the vertical distance (row offset) between the current
     * square and the next one, for all possible 12 moves.
     */
    int r_offset [] = { -2, -1,  1,  2,  2,  1, -1, -2,  3, -3,  0,  0};

    /**
     * Holds the horizontal distance (column offset) between the current
     * square and the next one, for all possible 12 moves.
     */
    int c_offset [] = {  1,  2,  2,  1, -1, -2, -2, -1,  0,  0, -3,  3};

    /**
     * Controls whether moves in the same row/column will be allowed.
     */
    boolean knightMovesOnly = true;

    /**
     * Number of allowed moves: 8 if knightMovesOnly is true, 12
     * otherwise.
     */
    int allowedMovesCount;

	/** Table for storing moves already tried out.*/
	boolean [][] lockedMovesArray;

    /** Row of first square when solving.*/
    int firstMoveRow;

    /** Column of first square when solving.*/
    int firstMoveColumn;

    /**
     * Table that stores mobility of squares (number of squares to which we
     * can move from each square).
     */
    int mobilityTable [];

	// Time variables
	Date startTime;
	Date endTime;

	/*void D(String s)
	{
		System.out.println(s);
	}*/


	class BoardCommand extends Command
	{
		int tag;

		BoardCommand(String label, int type, int pri, int tag_)
		{
			super(label, type, pri);
			tag = tag_;
		}
	}

	/**
	 * Initialises board.
	 */
    public void initBoard()
	{
    	resetBoard();

    	try
    	{
    		knightImage =
    			Image.createImage("/games/knightstour/images/knight.png");
		}
		catch(IOException e){}
    }

	/**
	 * Resets board's variables.
	 */
    public void resetBoard()
	{
        movesTableSize = boardSize*boardSize;
        allowedMovesCount = (knightMovesOnly) ? 8 : 12;

        mobilityTable = new int [movesTableSize];

        currentPosition = -1;
        currentValue = 0;

		if (movesTableSize > 0)
		{
			movesTable = new int[movesTableSize];
			for(int i =0; i<movesTableSize; i++)
				movesTable[i] = -1;
		}

		selPos = 0;
		startTime = new Date();
    }

	/**
	 * Sets game's state. Updates commands list.
	 */
	void setState(int ns)
	{
		gameState = ns;
		removeCommand(cmd[CMD_UNDO]);
		removeCommand(cmd[CMD_RESET]);
		removeCommand(cmd[CMD_BEST_TIME]);
		removeCommand(cmd[CMD_ABOUT]);
		removeCommand(cmd[CMD_HELP]);
		removeCommand(cmd[CMD_EXIT]);
		switch (gameState) {
			case START:
				addCommand(cmd[CMD_BEST_TIME]);
				addCommand(cmd[CMD_ABOUT]);
				addCommand(cmd[CMD_HELP]);
				addCommand(cmd[CMD_EXIT]);
				break;
			case PLAYING:
				addCommand(cmd[CMD_UNDO]);
				addCommand(cmd[CMD_RESET]);
				addCommand(cmd[CMD_BEST_TIME]);
				addCommand(cmd[CMD_ABOUT]);
				addCommand(cmd[CMD_HELP]);
				addCommand(cmd[CMD_EXIT]);
				break;
			case WON:
			case ENDED:
				addCommand(cmd[CMD_RESET]);
				addCommand(cmd[CMD_BEST_TIME]);
				addCommand(cmd[CMD_ABOUT]);
				addCommand(cmd[CMD_HELP]);
				addCommand(cmd[CMD_EXIT]);
				break;
		}
	}

	/**
	 * Stores midlet, display variables.
	 * Set up commands, initialises board.
	 */
	public Board(MIDlet midlet_)
	{
		int i;

		// "global" variables
		midlet = midlet_;
		dpy = Display.getDisplay(midlet);

		// set up commands
		cmd = new Command[CMD_ZLAST];

		cmd[CMD_UNDO] =
			new BoardCommand("Undo", Command.SCREEN, 1, CMD_UNDO);

		cmd[CMD_BEST_TIME] =
			new BoardCommand("Best Time", Command.SCREEN, 2, CMD_BEST_TIME);

		cmd[CMD_RESET] =
			new BoardCommand("Reset", Command.SCREEN, 1, CMD_RESET);

		cmd[CMD_ABOUT] =
			new BoardCommand("About", Command.HELP, 3, CMD_ABOUT);

		cmd[CMD_HELP] =
			new BoardCommand("Help", Command.HELP, 3, CMD_HELP);

		cmd[CMD_EXIT] =
			new BoardCommand("Exit", Command.EXIT, 4, CMD_EXIT);

		// set up the listener
		setCommandListener(this);

		// set up initial state
		setState(PLAYING);

		// Init Board
		initBoard();
	}

	/**
	 * Command actions handling.
	 */
	public void commandAction(Command c, Displayable d) {
		switch (((BoardCommand) c).tag) {
			case CMD_ABOUT:
				About.showAbout(Display.getDisplay(midlet));
				break;
			case CMD_HELP:
				//HelpSolo.showHelp(Display.getDisplay(midlet));
				break;
			case CMD_EXIT:
				midlet.notifyDestroyed();
				break;
			case CMD_RESET:
				setState(PLAYING);
				resetBoard();
				repaint();
				break;
			case CMD_UNDO:
				undoMove();
				repaint();
				break;
			case CMD_BEST_TIME:
				long bestTime = 0;
				Score score = new Score(1);
				if (score.open("KnightTourScores"))
				{
					bestTime = score.getTime(0);
				}
				score.close();
				long sec = (bestTime / 1000) % 60,
					 min = (bestTime / 1000) /60;
				String str = new String();
				if (bestTime == 0)
					str = "Not solved";
				else if (sec >= 10)
					str +=  min + ":" + sec;
				else
					str +=  min + ":0" + sec;

				Alert alert = new Alert("Knight Tour");
				alert.setTimeout(Alert.FOREVER);
				alert.setString(str);
				dpy.setCurrent(alert);
				break;
		}
	}

	/**
	 * Paints the puzzle's board.
	 *
	 * @param g Graphics object.
	 */
    public void paint(Graphics g)
    {
        String str;
  		g.setColor(0xFFFFFF);
  		g.fillRect(0, 0, getWidth(), getHeight());

		g.setColor(0);
		Font oldFont = g.getFont(),
			 newFont;
		boolean showNumbers = false;

		try
		{
			newFont = Font.getFont(Font.FACE_SYSTEM,
									Font.STYLE_PLAIN,
									Font.SIZE_SMALL);
			g.setFont(newFont);

			if (newFont.charWidth('5') < 6)
				showNumbers = true;
		}
		catch (IllegalArgumentException e)
		{
		}


        for(int r = 0; r < boardSize; r++)
        {
            for(int c = 0; c < boardSize; c++)
            {
                // Draw Rect
                g.drawRect(X_OFFSET+c*TILE_SIZE,
                        Y_OFFSET+r*TILE_SIZE,
                        TILE_SIZE,
                        TILE_SIZE);

                // Draw highlighted rect
                if (currentPosition != -1 &&
                    isMoveAllowed(r, c))
                {
                    g.drawRect(X_OFFSET+c*TILE_SIZE + 1,
                            Y_OFFSET+r*TILE_SIZE + 1,
                            TILE_SIZE-2,
                            TILE_SIZE-2);
                }

                // Draw numbers
                if (movesTable[r*boardSize+c]!=-1)
                {
                    int val = movesTable[r*boardSize+c] + 1;
                    str = "" + val;

                    if (r*boardSize+c != currentPosition)
                    {
						if (showNumbers)
						{
							g.drawString(str,
									X_OFFSET+c*TILE_SIZE+TILE_SIZE/2 + 1,
									Y_OFFSET+r*TILE_SIZE + 1,
									Graphics.TOP | Graphics.HCENTER);
						}
						else
						{
							g.setColor(0x808080);
							g.fillRect(X_OFFSET + c*TILE_SIZE + 1,
									   Y_OFFSET+r*TILE_SIZE + 1,
									   TILE_SIZE - 1,
									   TILE_SIZE - 1);
							g.setColor(0x000000);
						}
					}
					else
					{
						g.drawImage(knightImage,
								X_OFFSET + c*TILE_SIZE + 1,
								Y_OFFSET + r*TILE_SIZE,
								0);
					}
                }
            }
        }
        g.setFont(oldFont);

        // Hightlight selected position
        if (selPos >= 0)
        {
            int r = selPos/boardSize;
            int c = selPos%boardSize;

            g.setColor(0xFF0000);
            g.drawRect(X_OFFSET+c*TILE_SIZE,
                    X_OFFSET+r*TILE_SIZE,
                    TILE_SIZE,
                    TILE_SIZE);
            g.drawRect(X_OFFSET+c*TILE_SIZE + 1,
                    X_OFFSET+r*TILE_SIZE + 1,
                    TILE_SIZE-2,
                    TILE_SIZE-2);
            g.setColor(0x000000);
        }

		if (gameState == WON)
		{
			Alert alert = new Alert("Well done!");
			alert.setTimeout(Alert.FOREVER);

			long time = endTime.getTime() - startTime.getTime();
			long bestTime = 0;

			// Read best time, update if necessary
			Score score = new Score(1);
			if (score.open("KnightTourScores"))
			{
				bestTime = score.getTime(0);
				if (bestTime > time || bestTime == 0)
					score.setTime(0, (int) time);
			}
			score.close();

			long sec = (time / 1000) % 60,
				 min = (time / 1000) /60;

			String strScore = new String();
			if (bestTime > time || bestTime == 0)
				strScore = "New best time!\n";
			else
				strScore = "Solved\nTime: ";
			if (sec >= 10)
				strScore +=  min + ":" + sec;
			else
				strScore +=  min + ":0" + sec;
			alert.setString(strScore);

			Display.getDisplay(midlet).setCurrent(alert);
			gameState = ENDED;
		}
    }

    private void selectSquare(Graphics g,
                              int r,
                              int c)
    {
		int oldColor = g.getColor();

		g.setColor(0xFF0000);
		g.drawRect(X_OFFSET - 1 + c*TILE_SIZE,
				   Y_OFFSET - 1 + r*TILE_SIZE,
				   TILE_SIZE + 1,
				   TILE_SIZE + 1);
		g.setColor(oldColor);
	}


	/**
	 * Key Pressed handling
	 */
	public void keyPressed(int code)
	{
		if (gameState != PLAYING)
			return;

		int game = getGameAction(code);

		switch (game)
		{
			case Canvas.UP:
				if (selPos > 7)
					selPos-=8;
				else
					selPos+=56;
				break;
			case Canvas.DOWN:
				if (selPos < 56)
					selPos+=8;
				else
					selPos-=56;
				break;
			case Canvas.LEFT:
				if (selPos % 8 == 0)
					selPos+=7;
				else
					selPos--;
				break;
			case Canvas.RIGHT:
				selPos++;
				if (selPos % 8 == 0)
					selPos-=8;
				break;
			case Canvas.FIRE:
				int r = selPos / 8,
				    c = selPos % 8;

				makeMove(r, c);
				break;
			default:
			return;
		}

		repaint();

		if (isSolved())
		{
			endTime = new Date();
			setState(WON);
		}
	}

    /**
     * Makes the move at the selected square.
     *
     * @param r Row of selected square.
     * @param c Column of selected square.
     */
    public void makeMove(int r, int c)
    {
        if (c < boardSize && r < boardSize &&
            c >= 0 && r >= 0)
        {
            if (isMoveAllowed(r, c))
            {
                movesTable[c+r*boardSize] = currentValue++;
                currentPosition = c+r*boardSize;

                if (currentValue == movesTableSize)
                    gameState = WON;
            }
            else if (movesTable[c+r*boardSize] == currentValue-1)
            {
                // Undo move
                movesTable[c+r*boardSize] = -1;
                currentValue--;
                currentPosition = -1;
                if (currentValue > 0)
                    for(int i=0; i<movesTableSize; i++)
                        if (movesTable[i] == currentValue-1)
                        {
                            currentPosition = i;
                            break;
                        }
            }
        }
    }

    /**
     * Undoes last move.
     */
    public void undoLastMove()
    {
        if (currentValue > 0)
        {
            makeMove(currentPosition / boardSize,
                     currentPosition % boardSize);
        }
    }

    /**
     * Checks if the move at the selected square is allowed.
     *
     * @param r Row of selected square.
     * @param c Column of selected square.
     * @return true is move is allowed.
     */
    private boolean isMoveAllowed(int r, int c)
    {
        if (c >= boardSize || r >= boardSize ||
            c < 0 || r < 0)
            return(false);

        if (movesTable[c+r*boardSize] != -1)
            return(false);

        if (currentPosition == -1)
            return(true);

        int currentPosition_r = currentPosition / boardSize;
        int currentPosition_c = currentPosition % boardSize;

        for(int i=0; i<allowedMovesCount; i++)
        {
            if (r == currentPosition_r + r_offset[i] &&
                c ==  currentPosition_c + c_offset[i] &&
                movesTable[r*boardSize + c] == -1)
            {
                return(true);
            }
        }

        return(false);
    }

	/**
	 * Puzzle's routine. Undoes last move
	 */
    public void undoMove()
	{
        if (currentValue > 0)
        {
            makeMove(currentPosition / boardSize,
                     currentPosition % boardSize);
        }
	}

	/**
	 * Puzzle's routine. Returns true if the puzzle is solved.
	 */
	boolean isSolved()
	{
		return(currentValue == movesTableSize);
	}
}
